@@ -1,6 +1,6 @@
 from threading import Lock
 from time import sleep, time
-from math import ceil
+from math import ceil, inf
 from os import environ
 
 from psutil import Process, cpu_count, cpu_times
@@ -158,7 +158,7 @@ def wait_for_idle_cpus(
             # Precalculate the quota with headroom for a new job
             cpu_quota -= required
         else:
-            cpu_quota = None
+            cpu_quota = inf
             if not max_concurrent:
                 max_concurrent = ceil(
                     cpu_count() * DEFAULT_MAX_CONCURRENT_MULTIPLIER / required
@@ -174,7 +174,7 @@ def wait_for_idle_cpus(
         if idle - reserved < required:
             return False
 
-        return cpu_quota is None or cpu_used <= cpu_quota
+        return cpu_used <= cpu_quota
 
     def wait_for_slot(halt):
         start_time = time()
@@ -189,12 +189,10 @@ def wait_for_idle_cpus(
             ),
         )
         while len(worker_list) and not halt():
-            reserved = worker_list.reserved_cpu_amount()
             idle_deriv = idle.update()
-            if (
-                len(worker_list) < max_concurrent
-                and idle_deriv - reserved >= required
-                and (cpu_quota is None or cpu_used.update() + reserved <= cpu_quota)
+            cpu_used_deriv = cpu_used.update()
+            if len(worker_list) < max_concurrent and cpu_is_available(
+                idle_deriv, cpu_used_deriv
             ):
                 break
 
