@@ -26,7 +26,7 @@ from werkzeug.serving import WSGIRequestHandler
 from .gating import wait_for_idle_cpus
 from .oob_response import OutOfBandResponder
 from .resource_balancer import BoundedResourceBalancer, ResourceLoan
-from .util import ObjectProxy
+from .util import ObjectProxy, ThreadsTracker
 
 LOGGER = getLogger("pothead.worker")
 BROKER_CONNECT_RETRY_INTERVAL = 2
@@ -162,13 +162,14 @@ class Server:
         self.broker_balancer = load_balancer(host, port)
         self.sockets_waiting_for_request: Set[socket] = set()
         self.real_app = app
+        self.worker_tracker = ThreadsTracker()
         self.app = (
             OutOfBandResponder(app, redirect_port) if redirect_port is not None else app
         )
 
     def worker(self):
         while self.running:
-            with self.broker_balancer.acquire_addr() as addr:
+            with self.worker_tracker, self.broker_balancer.acquire_addr() as addr:
                 self._fetch_and_run_one_job(addr)
 
     def _fetch_and_run_one_job(
@@ -201,7 +202,7 @@ class Server:
         return self.app
 
     def shutdown(self):
-        LOGGER.info("shutting down workers...")
+        LOGGER.info("Shutting down workers...")
         self.running = False
         for s in list(self.sockets_waiting_for_request):
             try:
@@ -210,6 +211,12 @@ class Server:
             except OSError:
                 pass
 
+        self.worker_tracker.drain()
+
+        if isinstance(self.app, OutOfBandResponder):
+            self.app.shutdown()
+        LOGGER.info("Worker shutdown completed")
+
     def poll_loop(self):
         """Uses app.wait_for_slot() spawn new workers dynamically. This is useful, for example to wait for
         enough CPU, memory, or disk space, to start next job.
@@ -230,6 +237,8 @@ class Server:
 
             # This will block as long as we are waiting for requests from all broker-instances
             broker_instance_lease = self.broker_balancer.acquire_addr()
+            if not self.running:
+                break
 
             Thread(
                 target=self._run_one_poll, args=(broker_instance_lease, callbacks)
@@ -240,7 +249,7 @@ class Server:
             broker_instance_lease.release()
             getattr(callbacks, "request_received", lambda _: None)(environ)
 
-        with broker_instance_lease as addr:
+        with self.worker_tracker, broker_instance_lease as addr:
             self._fetch_and_run_one_job(
                 addr, request_received, getattr(callbacks, "done", lambda _: None)
             )
@@ -253,7 +262,7 @@ def install_term_handler(f):
     previous = []
 
     def on_term(_signal, _stack):
-        f()
+        Thread(target=f).start()
         for p in previous:
             p()
 
@@ -332,6 +341,7 @@ if __name__ == "__main__":
             "Starting up job-polling dynamic workers connecting to %s", args.connect
         )
         server.poll_loop()
+        LOGGER.info("connect-loop shut down completed")
     else:
         LOGGER.info(
             "Starting up %d workers connecting to %s", args.workers, args.connect
@@ -341,5 +351,3 @@ if __name__ == "__main__":
             worker.start()
         for worker in workers:
             worker.join()
-
-    LOGGER.info("shut down")
